@@ -278,7 +278,7 @@ class CircuitOperationList(list):
             if self.table[op_name].arity:
                 self._set_specific_operation_shortcut(op_name)
 
-    def __getitem__(self, item) -> CircuitOperationList | CircuitOperation:
+    def __getitem__(self, item) -> CircuitOperationList | CircuitOperation:  # type: ignore[override]  # type: ignore[override]  # type: ignore[override]
         """For the builtin list, this method is used both for accessing a single element: ``mylist[0]`` and accessing
         a slice: ``mylist[1:3]``. The latter should generate a new CircuitOperationList, so we override the method to
         ensure that it does.
@@ -350,11 +350,6 @@ class CircuitOperationList(list):
         qubit_names = self.qubits
         if name not in self.table:
             raise KeyError(f"QuantumOp with name {name} is not in the gate definitions table.")
-        arity = self.table[name].arity
-        if len(locus_indices) != len(set(locus_indices)):
-            raise ValueError("Repeated locus indices.")
-        if arity and arity != len(locus_indices):  # arity = 0 is barrier and measure
-            raise ValueError(f"Operation {name} has {arity=} but {len(locus_indices)} target qubits were provided.")
 
         try:
             locus = tuple(qubit_names[idx] for idx in locus_indices)
