@@ -18,7 +18,7 @@ The reset operation is a non-unitary quantum channel that sets the state of a qu
 
 from __future__ import annotations
 
-from collections.abc import Iterable
+from collections.abc import Iterable, Mapping
 
 from exa.common.data.parameter import Parameter
 from exa.common.qcm_data.chip_topology import ChipTopology
@@ -53,7 +53,7 @@ class Reset_Conditional(CompositeGate):
     blocked.
     """
 
-    registered_gates = ["measure"]
+    registered_gates = ("measure", "cc_prx")
 
     def _call(self) -> TimeBox:  # type: ignore[override]
         # find locus components that are resettable via conditional reset
@@ -64,11 +64,11 @@ class Reset_Conditional(CompositeGate):
         )
         # try to get a qnd measurement, otherwise use default one
         try:  # TODO: should the QND measurement be its own QuantumOp instead?
-            probe_timebox = self.build("measure", resettable, impl_name="constant_qnd").probe_timebox(
+            probe_timebox = self.build("measure", resettable, impl_name="constant_qnd").probe_timebox(  # type: ignore[attr-defined]
                 RESET_MEASUREMENT_KEY, feedback_key=RESET_FEEDBACK_KEY
             )
         except (ValueError, KeyError):
-            probe_timebox = self.build("measure", resettable).probe_timebox(
+            probe_timebox = self.build("measure", resettable).probe_timebox(  # type: ignore[attr-defined]
                 RESET_MEASUREMENT_KEY, feedback_key=RESET_FEEDBACK_KEY
             )
         virtual_channels = set()
@@ -89,7 +89,7 @@ class Reset_Conditional(CompositeGate):
         probe_timebox.neighborhood_components[0] = blocks
         measure = TimeBox.composite([probe_timebox])
         measure.neighborhood_components[0] = blocks
-        return TimeBox.composite([measure, resets])
+        return TimeBox.composite([measure, resets])  # type: ignore[list-item]
 
     @classmethod
     def get_locus_mapping_name(cls, operation_name: str, implementation_name: str) -> str:
@@ -118,17 +118,10 @@ class Reset_Wait(GateImplementation):
         Does not block any additional components beside the locus itself.
         """
         if len(self.locus) == 1:
-            waits_box = self.builder.wait(self.locus, self.calibration_data["duration"])
+            waits_box = self.builder.wait(self.locus, self.calibration_data["duration"], rounding=True)
         else:
-            prio_calibration = self.calibration_data if self.calibration_data else None
-            waits_box = TimeBox.composite(
-                [
-                    self.builder.get_implementation(
-                        self.parent.name, (q,), impl_name=self.name, priority_calibration=prio_calibration
-                    ).wait_box()
-                    for q in self.locus
-                ]
-            )
+            # factorizability: use the sub-implementations
+            waits_box = TimeBox.composite([self.sub_implementations[c].wait_box() for c in self.locus])  # type: ignore[attr-defined]
         return waits_box
 
     def _call(self) -> TimeBox:
@@ -147,14 +140,13 @@ class Reset_Wait(GateImplementation):
         return TimeBox.composite([waits_box])
 
     def duration_in_seconds(self) -> float:
-        return max(
-            self.builder.get_implementation(self.parent.name, (q,), impl_name=self.name).calibration_data["duration"]
-            for q in self.locus
-        )
+        if len(self.locus) == 1:
+            return self.calibration_data["duration"]
+        return max(self.sub_implementations[c].calibration_data["duration"] for c in self.locus)
 
     @classmethod
     def get_custom_locus_mapping(
-        cls, chip_topology: ChipTopology, component_to_channels: dict[str, Iterable[str]]
+        cls, chip_topology: ChipTopology, component_to_channels: Mapping[str, Iterable[str]]
     ) -> dict[tuple[str, ...] | frozenset[str], tuple[str, ...]] | None:
         """Supported loci: all components that have channels."""
         return {(c,): (c,) for c in chip_topology.all_components}
